import os
import yaml
import time
from dotenv import load_dotenv
from typing import Dict, List, Optional, Union
import threading
import logging
from pydantic import (
    BaseModel,
    Field,
    HttpUrl,
    ValidationError,
    model_validator,
    ConfigDict,
)


# Initialize logger
logger = logging.getLogger(__name__)


# Load environment variables from .env file
load_dotenv()


# Helper function to parse API keys from various formats
def _parse_api_keys(api_key_input: Union[str, List[str]]) -> List[str]:
    """
    Parse API keys from various input formats:
    - Single string: "key1" -> ["key1"]
    - Comma-separated string: "key1,key2,key3" -> ["key1", "key2", "key3"]
    - List of strings: ["key1", "key2"] -> ["key1", "key2"]
    """
    if isinstance(api_key_input, list):
        # Filter out empty strings and strip whitespace
        return [key.strip() for key in api_key_input if key.strip()]
    elif isinstance(api_key_input, str):
        if "," in api_key_input:
            # Comma-separated values
            return [key.strip() for key in api_key_input.split(",") if key.strip()]
        else:
            # Single key
            return [api_key_input.strip()] if api_key_input.strip() else []
    else:
        return []


# Define built-in providers with their base URLs and corresponding environment variable prefixes
BUILTIN_PROVIDERS = {
    "openai": {
        "base_url": "https://api.openai.com/v1",
        "api_key_env": "OPENAI_API_KEY",
        "default_model": "gpt-4o",
        "max_tokens": 4096,
    },
    "anthropic": {
        "base_url": "https://api.anthropic.com/v1",
        "api_key_env": "ANTHROPIC_API_KEY",
        "default_model": "claude-3-5-sonnet-20240620",
        "max_tokens": 4096,
    },
    "openrouter": {
        "base_url": "https://openrouter.ai/api/v1",
        "api_key_env": "OPENROUTER_API_KEY",
        "default_model": "mistralai/mistral-7b-instruct",
        "max_tokens": 4096,
    },
    "togetherai": {
        "base_url": "https://api.together.xyz/v1",
        "api_key_env": "TOGETHER_API_KEY",
        "default_model": "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "max_tokens": 4096,
    },
    "perplexity": {
        "base_url": "https://api.perplexity.ai",
        "api_key_env": "PERPLEXITY_API_KEY",
        "default_model": "llama-3-sonar-small-32k-online",
        "max_tokens": 32768,
    },
    "mistralai": {
        "base_url": "https://api.mistral.ai/v1",
        "api_key_env": "MISTRAL_API_KEY",
        "default_model": "mistral-large-latest",
        "max_tokens": 32768,
    },
    "fireworks": {
        "base_url": "https://api.fireworks.ai/inference/v1",  # Note: Fireworks specific endpoint often ends with /v1 or /v1/chat/completions
        "api_key_env": "FIREWORKS_API_KEY",
        "default_model": "accounts/fireworks/models/mixtral-8x7b-instruct",
        "max_tokens": 32768,
    },
    "groq": {
        "base_url": "https://api.groq.com/openai/v1",
        "api_key_env": "GROQ_API_KEY",
        "default_model": "llama3-8b-8192",
        "max_tokens": 32768,
    },
    "deepinfra": {
        "base_url": "https://api.deepinfra.com/v1",
        "api_key_env": "DEEPINFRA_API_KEY",
        "default_model": "mistralai/Mistral-7B-Instruct-v0.2",
        "max_tokens": 32768,
    },
    "anyscale": {
        "base_url": "https://api.endpoints.anyscale.com/v1",
        "api_key_env": "ANYSCALE_API_KEY",
        "default_model": "meta-llama/Llama-2-7b-chat-hf",
        "max_tokens": 4096,
    },
    "novita": {
        "base_url": "https://api.novita.ai/v1",  # Based on common OpenAI-compatible patterns, though documentation might specify /v1/chat/completions
        "api_key_env": "NOVITA_API_KEY",
        "default_model": "llama2-7b-chat",
        "max_tokens": 8192,
    },
    "cerebras": {
        "base_url": "https://api.cerebras.ai/v1",
        "api_key_env": "CEREBRAS_API_KEY",
        "default_model": "llama3.1-8b",
        "max_tokens": 2048,
    },
    "featherless": {
        "base_url": "https://api.featherless.ai/v1",
        "api_key_env": "FEATHERLESS_API_KEY",
        "default_model": "meta-llama/Meta-Llama-3.1-8B-Instruct",
        "max_tokens": 8192,
    },
    "cohere": {
        "base_url": "https://api.cohere.ai/compatibility/v1",
        "api_key_env": "COHERE_API_KEY",
        "default_model": "command-r-plus",
        "max_tokens": 131072,
    },
    "qwen": {
        "base_url": "https://dashscope-intl.aliyuncs.com/compatible-mode/v1",
        "api_key_env": "DASHSCOPE_API_KEY",
        "default_model": "qwen-plus",
        "max_tokens": 32768,
    },
    "deepseek": {
        "base_url": "https://api.deepseek.com/v1",
        "api_key_env": "DEEPSEEK_API_KEY",
        "default_model": "deepseek-chat",
    },
    "google": {
        "base_url": "https://generativelanguage.googleapis.com/v1beta/openai/",
        "api_key_env": "GEMINI_API_KEY",
        "default_model": "gemini-2.5-flash",
        "max_tokens": 32768,
    },
    # Add other built-in providers here as needed
}


class LLMProviderConfig(BaseModel):
    name: str
    base_url: HttpUrl
    model: str
    api_key: str  # This will be populated with the current round-robin key
    temperature: float = 0.7
    max_tokens: int = Field(..., gt=0)

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    def __init__(self, **data):
        super().__init__(**data)
        # Ensure _api_keys is properly initialized
        if not hasattr(self, "_api_keys") or not self._api_keys:
            self._api_keys: List[str] = [self.api_key] if self.api_key else []
        if not hasattr(self, "_current_key_index"):
            self._current_key_index: int = 0

    @model_validator(mode="before")
    @classmethod
    def validate_api_key(cls, data):
        if isinstance(data, dict):
            api_keys_value = data.get("api_keys")
            api_key_value = data.get("api_key")

            if api_keys_value is not None:
                # api_keys takes precedence
                api_keys = _parse_api_keys(api_keys_value)
                if api_keys:
                    data["api_key"] = api_keys[0]  # Set first key as default
                    data["_api_keys"] = api_keys  # Store all keys
            elif api_key_value is not None:
                # Single api_key, check if it's comma-separated
                api_keys = _parse_api_keys(api_key_value)
                data["api_key"] = (
                    api_keys[0] if api_keys else api_key_value
                )  # Set first key as default
                data["_api_keys"] = api_keys  # Store all keys (even if just one)

            # Remove original api_keys field from data to avoid Pydantic validation issues
            if "api_keys" in data:
                del data["api_keys"]
        return data

    def get_next_api_key(self) -> str:
        """Get the next API key in round-robin fashion."""
        if not self._api_keys:
            return self.api_key

        # Get current key
        current_key = self._api_keys[self._current_key_index]

        # Move to next key for next call
        self._current_key_index = (self._current_key_index + 1) % len(self._api_keys)

        # Update the api_key field for compatibility
        self.api_key = current_key

        return current_key

    def set_api_keys(self, api_keys: List[str]):
        """Set the list of API keys for round-robin."""
        if api_keys:
            self._api_keys = api_keys
            self._current_key_index = 0
            self.api_key = api_keys[0]

    def has_multiple_keys(self) -> bool:
        """Check if this provider has multiple API keys."""
        return len(self._api_keys) > 1


# Global cache for built-in providers to share state across all BorgLLM instances
_GLOBAL_BUILTIN_PROVIDERS: Dict[str, LLMProviderConfig] = {}
_GLOBAL_BUILTIN_LOCK = threading.Lock()


class VirtualLLMProviderConfig(BaseModel):
    name: str
    upstreams: List[Dict[str, str]]


class LLMConfig(BaseModel):
    providers: List[LLMProviderConfig]
    virtual: Optional[List[VirtualLLMProviderConfig]] = None
    default_model: Optional[str] = None


class BorgLLM:
    _instance = None
    _config_initialized: bool = False
    _lock = threading.Lock()  # Use a lock for thread-safe singleton initialization

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                # Double-check inside lock
                if cls._instance is None:
                    cls._instance = super(BorgLLM, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        config_path: str = "borg.yaml",
        initial_config_data: Optional[Dict] = None,
        _force_reinitialize: bool = False,  # Internal flag for testing/reinitialization
    ):
        # Initialize instance variables if not already present or if reinitialization is forced.
        # This prevents _default_provider_name from being wiped out on subsequent __init__ calls
        # when the singleton instance itself hasn't been destroyed.
        if not hasattr(self, "_default_provider_name") or _force_reinitialize:
            self._default_provider_name: Optional[str] = None
        if not hasattr(self, "_config") or _force_reinitialize:
            self._config: Optional[LLMConfig] = None
        if not hasattr(self, "_real_providers") or _force_reinitialize:
            self._real_providers: Dict[str, LLMProviderConfig] = {}
        if not hasattr(self, "_virtual_providers") or _force_reinitialize:
            self._virtual_providers: Dict[str, VirtualLLMProviderConfig] = {}
        if not hasattr(self, "_unusable_providers") or _force_reinitialize:
            self._unusable_providers: Dict[str, float] = (
                {}
            )  # provider_name -> cooldown_end_timestamp
        if not hasattr(self, "_virtual_provider_last_index") or _force_reinitialize:
            self._virtual_provider_last_index: Dict[str, int] = {}

        if BorgLLM._config_initialized and not _force_reinitialize:
            logger.debug("BorgLLM already initialized, skipping re-initialization.")
            return

        with self._lock:  # Ensure thread-safe initialization
            if not BorgLLM._config_initialized or _force_reinitialize:
                self.config_path = config_path
                self._config_paths = self._get_config_paths(config_path)
                self._config: Optional[LLMConfig] = None
                self._real_providers: Dict[str, LLMProviderConfig] = {}
                self._virtual_providers: Dict[str, VirtualLLMProviderConfig] = {}
                # self._config_loaded: bool = False # This can be inferred from _config_initialized

                # New instance-specific default provider
                self._default_provider_name: Optional[str] = None

                if initial_config_data:
                    # Interpolate environment variables for initial_config_data as well
                    interpolated_data = self._interpolate_env_variables(
                        initial_config_data
                    )

                    # Process providers to handle api_keys field
                    if (
                        "llm" in interpolated_data
                        and "providers" in interpolated_data["llm"]
                    ):
                        for provider_config in interpolated_data["llm"]["providers"]:
                            self._process_provider_api_keys(provider_config)

                    try:
                        self._config = LLMConfig(**interpolated_data["llm"])
                        pass
                    except ValidationError as e:
                        logger.error(f"Initial configuration validation error: {e}")
                        raise

                self._load_config()  # Always attempt to load from file after initial_config_data

                self._populate_providers()
                self._add_builtin_providers()

                # Log loaded API keys summary
                logger.info("\n--- BorgLLM Configuration Summary ---")
                if self._real_providers:
                    for provider_name, provider_config in self._real_providers.items():
                        masked_key = (
                            f"{provider_config.api_key[:4]}...{provider_config.api_key[-4:]}"
                            if provider_config.api_key
                            else "[NO KEY]"
                        )
                        num_keys = (
                            len(provider_config._api_keys)
                            if provider_config._api_keys
                            else 1
                        )
                        logger.info(
                            f"Provider '{provider_name}': {num_keys} API key(s) loaded (Current/First: {masked_key})"
                        )
                else:
                    logger.info(
                        "No LLM providers configured or found via environment variables."
                    )
                logger.info("-----------------------------------")

                self._virtual_provider_last_index: Dict[str, int] = {}
                self._unusable_providers: Dict[str, float] = {}
                # self._lock = threading.Lock() # Now a class-level lock for singleton

                BorgLLM._config_initialized = True

    @classmethod
    def get_instance(
        cls,
        config_path: str = "borg.yaml",
        initial_config_data: Optional[Dict] = None,
    ):
        """
        Provides access to the singleton BorgLLM instance.
        If the instance has not been initialized, it will be initialized with the provided
        config_path and initial_config_data.
        """
        if cls._instance is None:
            # This path is taken only if __new__ hasn't created it yet (e.g., direct call to get_instance)
            # or if it needs to be re-initialized for testing.
            cls(
                _force_reinitialize=True,
                config_path=config_path,
                initial_config_data=initial_config_data,
            )
        elif (
            not cls._config_initialized
        ):  # Handle cases where instance exists but wasn't fully initialized
            cls(
                _force_reinitialize=True,
                config_path=config_path,
                initial_config_data=initial_config_data,
            )
        return cls._instance

    @property
    def config(self) -> Optional[LLMConfig]:
        """Public property to access the configuration."""
        return self._config

    @property
    def providers(self) -> Dict[str, LLMProviderConfig]:
        """Public property to access the real providers."""
        return self._real_providers

    def set_default_provider(self, provider_name: str):
        """
        Sets the default LLM provider name for this BorgLLM instance.

        This method sets the default provider that BorgLLM will use when no provider
        is explicitly specified in a call to get(). This setting takes precedence
        over any default_model specified in the borg.yaml configuration file.

        Args:
            provider_name: The name of the provider to set as the default.
        """
        # Check if it's a built-in provider reference (e.g., "google:gemini-2.5-flash")
        is_builtin_reference = False
        if ":" in provider_name:
            provider_key = provider_name.split(":", 1)[0]
            if provider_key in BUILTIN_PROVIDERS:
                is_builtin_reference = True
        elif provider_name in BUILTIN_PROVIDERS:
            is_builtin_reference = True

        # Validate that the provider exists
        if (
            provider_name not in self._real_providers
            and provider_name not in self._virtual_providers
            and not is_builtin_reference
        ):
            raise ValueError(
                f"Provider '{provider_name}' not found. Cannot set as default."
            )

        with self._lock:
            self._default_provider_name = provider_name
            logger.info(
                f"Instance default LLM provider set to '{provider_name}' (overrides any config file default)."
            )

    def _get_config_paths(self, base_path: str) -> List[str]:
        # If the path explicitly ends with .yaml or .yml, use it directly
        if base_path.endswith((".yaml", ".yml")):
            return [base_path]

        # Otherwise, try both .yaml and .yml
        return [f"{base_path}.yaml", f"{base_path}.yml"]

    def _populate_providers(self):
        if not self._config:
            self._config = LLMConfig(providers=[], virtual=[], default_model=None)

        for provider in self._config.providers:
            # The _multiple_api_keys should now be set by the model_validator,
            # so we just ensure the Pydantic model itself stores it correctly.
            self._real_providers[provider.name] = provider

        if self._config.virtual:
            for provider in self._config.virtual:
                self._virtual_providers[provider.name] = provider

        # Set the default provider name based on the config's default_model
        if self._config.default_model and not self._default_provider_name:
            self._default_provider_name = self._config.default_model

        # Static check: Verify all upstreams in virtual providers exist
        if self._config.virtual:
            for (
                virtual_provider_name,
                virtual_config,
            ) in self._virtual_providers.items():
                for upstream_info in virtual_config.upstreams:
                    upstream_name = upstream_info["name"]

                    # Check if it's a built-in provider reference (e.g., "cerebras:qwen-3-32b")
                    is_builtin_reference = False
                    if ":" in upstream_name:
                        provider_key = upstream_name.split(":", 1)[0]
                        if provider_key in BUILTIN_PROVIDERS:
                            is_builtin_reference = True
                    elif upstream_name in BUILTIN_PROVIDERS:
                        is_builtin_reference = True

                    if (
                        upstream_name not in self._real_providers
                        and upstream_name not in self._virtual_providers
                        and not is_builtin_reference
                    ):
                        raise ValueError(
                            f"Virtual provider '{virtual_provider_name}' references non-existent upstream '{upstream_name}'."
                        )

    def _process_provider_api_keys(self, provider_config: Dict):
        """
        Process a provider config to handle both api_key and api_keys fields.
        This function is now largely redundant as the LLMProviderConfig's model_validator
        handles this during Pydantic model creation.
        However, it's kept to ensure that the initial dictionary data
        is correctly formatted *before* passing to Pydantic if needed elsewhere.
        """
        api_keys_value = provider_config.get("api_keys")
        api_key_value = provider_config.get("api_key")

        # Determine which keys to use based on priority
        if api_keys_value is not None:
            # api_keys takes precedence
            api_keys = _parse_api_keys(api_keys_value)
            if api_keys:
                provider_config["api_key"] = api_keys[0]  # Set first key as default
                # Keep api_keys for model_validator to process
            else:
                provider_config["api_key"] = ""  # Explicitly set to empty if no keys
        elif api_key_value is not None:
            # Single api_key, check if it's comma-separated
            api_keys = _parse_api_keys(api_key_value)
            if api_keys:
                provider_config["api_key"] = api_keys[0]  # Set first key as default
                if len(api_keys) > 1:
                    # If multiple keys were parsed, set api_keys field for model_validator
                    provider_config["api_keys"] = api_keys
            else:
                provider_config["api_key"] = ""  # Explicitly set to empty if no keys
        else:
            provider_config["api_key"] = ""  # No API key found at all

        # Don't remove api_keys field - let model_validator handle it

    def _setup_provider_api_keys(self, llm_config: Dict):
        """
        This function is largely superseded by the LLMProviderConfig's model_validator
        which now handles setting _api_keys during model instantiation.
        It's kept as a placeholder but its logic is less critical now.
        """
        pass  # The logic is now within LLMProviderConfig.validate_api_key

    def _add_builtin_providers(self):
        # We need a lock here because _GLOBAL_BUILTIN_PROVIDERS is shared
        with _GLOBAL_BUILTIN_LOCK:
            for provider_name, settings in BUILTIN_PROVIDERS.items():
                # Support multiple API keys: check for both *_API_KEYS and *_API_KEY
                api_key_env = settings["api_key_env"]
                api_keys_env = (
                    api_key_env + "S"
                )  # e.g., OPENAI_API_KEY -> OPENAI_API_KEYS

                # Priority: *_API_KEYS takes precedence over *_API_KEY
                api_keys_value = os.getenv(api_keys_env)
                api_key_value = os.getenv(api_key_env)

                api_keys_list = []
                if api_keys_value:
                    api_keys_list = _parse_api_keys(api_keys_value)
                elif api_key_value:
                    api_keys_list = _parse_api_keys(api_key_value)

                if api_keys_list and provider_name not in self._real_providers:
                    # Only add if API key(s) are present and not already defined in config file
                    try:
                        # Prepare data for LLMProviderConfig, including _api_keys which model_validator will handle
                        provider_data = {
                            "name": provider_name,
                            "base_url": settings["base_url"],
                            "model": settings["default_model"],
                            "api_key": api_keys_list[
                                0
                            ],  # Pass first key directly for api_key field
                            "api_keys": api_keys_list,  # Pass all keys for model_validator to process
                            "temperature": settings.get("temperature", 0.7),
                            "max_tokens": settings.get("max_tokens", 4096),
                        }

                        builtin_config = LLMProviderConfig(**provider_data)

                        # Store in both global and local caches
                        _GLOBAL_BUILTIN_PROVIDERS[provider_name] = builtin_config
                        self._real_providers[provider_name] = builtin_config
                    except ValidationError as e:
                        logger.error(
                            f"Error validating built-in provider {provider_name}: {e}"
                        )
                        # Decide whether to raise or just skip
                        pass  # Skip invalid built-in configs

    def _interpolate_env_variables(self, data):
        if isinstance(data, dict):
            return {k: self._interpolate_env_variables(v) for k, v in data.items()}
        elif isinstance(data, list):
            return [self._interpolate_env_variables(elem) for elem in data]
        elif isinstance(data, str) and data.startswith("${") and data.endswith("}"):
            env_var_name = data[2:-1]
            return os.getenv(env_var_name, data)  # Return original if not found
        return data

    def _load_config(self):
        loaded = False
        for path in self._config_paths:
            if os.path.exists(path):
                logger.info(f"Loading configuration from {path}")
                with open(path, "r") as f:
                    raw_config = yaml.safe_load(f)

                # Interpolate environment variables before Pydantic validation
                interpolated_config = self._interpolate_env_variables(raw_config)

                # Process providers to handle api_keys field. This ensures the raw data
                # is correctly formatted before Pydantic receives it, especially for
                # initial config or merged config.
                if (
                    "llm" in interpolated_config
                    and "providers" in interpolated_config["llm"]
                ):
                    for provider_config in interpolated_config["llm"]["providers"]:
                        self._process_provider_api_keys(provider_config)

                try:
                    # If initial_config_data was provided, merge it with file config
                    if self._config:
                        # Merge logic: prioritize file config over initial_config_data if there are overlaps.
                        # For simplicity here, if file exists, it augments/overrides initial_config_data.
                        file_config_data = interpolated_config.get("llm", {})
                        current_providers = [
                            p.model_dump() for p in self._config.providers
                        ]
                        current_virtual = (
                            [v.model_dump() for v in self._config.virtual]
                            if self._config.virtual
                            else []
                        )

                        # Combine providers, ensuring uniqueness by name (file config overrides initial)
                        new_providers = file_config_data.get("providers", [])
                        all_providers = {p["name"]: p for p in current_providers}
                        for p in new_providers:
                            # Apply _process_provider_api_keys to new providers from file
                            self._process_provider_api_keys(p)
                            all_providers[p["name"]] = p

                        new_virtual = file_config_data.get("virtual", [])
                        all_virtual = {v["name"]: v for v in current_virtual}
                        for v in new_virtual:
                            all_virtual[v["name"]] = v

                        combined_config_data = {
                            "providers": list(all_providers.values()),
                            "virtual": list(all_virtual.values()),
                            "default_model": file_config_data.get(
                                "default_model", self._config.default_model
                            ),  # File config's default_model takes precedence
                        }
                        self._config = LLMConfig(**combined_config_data)
                    else:
                        self._config = LLMConfig(**interpolated_config["llm"])

                    # _setup_provider_api_keys is now handled by LLMProviderConfig's model_validator
                    loaded = True
                    break  # Break after loading the first found config file
                except KeyError:
                    logger.warning(f"Configuration file {path} is missing 'llm' key.")
                except ValidationError as e:
                    logger.error(f"Configuration validation error for {path}: {e}")
                    # Decide whether to raise or just skip - for now, we skip and try next path if any
                    pass
                except Exception as e:
                    logger.error(f"Error loading configuration from {path}: {e}")
                    pass

        self._config_loaded = loaded
        if not loaded and not self._config:
            logger.info(
                f"No configuration file found at {', '.join(self._config_paths)}. Proceeding with environment variables and defaults only, or initial data if provided."
            )
            self._config = LLMConfig(
                providers=[], virtual=[], default_model=None
            )  # Ensure _config is not None even if no file/initial data

    def signal_429(self, provider_name: str, duration: int = 300):
        with self._lock:
            self._unusable_providers[provider_name] = time.time() + duration

    def _is_provider_unusable(self, provider_name: str) -> bool:
        with self._lock:
            if provider_name in self._unusable_providers:
                cooldown_end = self._unusable_providers[provider_name]
                current_time = time.time()
                if current_time < cooldown_end:
                    return True
                else:
                    del self._unusable_providers[provider_name]  # Cooldown expired
            return False

    def get(
        self,
        name: Optional[str] = None,
        approximate_tokens: Optional[int] = None,
        timeout: Optional[float] = None,
        allow_await_cooldown: bool = True,
    ) -> LLMProviderConfig:
        # Default provider logic - prioritize programmatic setting over config file
        if name is None:
            if (
                self._default_provider_name
            ):  # Use instance-specific default FIRST (set via set_default_provider)
                name = self._default_provider_name
                logger.info(f"Using programmatically set default provider: {name}")
            elif (
                self._config and self._config.default_model
            ):  # Then check config file default
                name = self._config.default_model
                logger.info(f"Using default provider from config file: {name}")
            elif (
                self._real_providers
            ):  # Fallback to first available if no explicit default
                name = next(iter(self._real_providers))
                logger.info(
                    f"No default_model specified in loaded config or instance, using first available provider: {name}"
                )
            else:
                raise ValueError(
                    "No default LLM provider specified and no configuration file found, "
                    "and no built-in providers are available (missing API keys). "
                    "Please specify a provider name, set a default in borg.yaml, or use set_default_provider()."
                )

        # Existing logic for built-in providers and configured providers
        provider_key = None
        model_name_for_request = None

        if ":" in name:
            parts = name.split(":", 1)
            provider_key = parts[0]
            model_name_for_request = parts[1]

        # Check if this is a built-in provider request (has ":" and provider_key is in BUILTIN_PROVIDERS)
        if provider_key and provider_key in BUILTIN_PROVIDERS:
            # Use global cache for built-in providers to share state across all BorgLLM instances
            with _GLOBAL_BUILTIN_LOCK:
                # Check if this built-in provider already exists in global cache
                if provider_key in _GLOBAL_BUILTIN_PROVIDERS:
                    provider_instance = _GLOBAL_BUILTIN_PROVIDERS[provider_key]

                    # Ensure the model is updated if a specific model was requested (e.g., 'openai:gpt-3.5-turbo')
                    if (
                        model_name_for_request
                        and provider_instance.model != model_name_for_request
                    ):
                        provider_instance.model = model_name_for_request

                    # Handle rate limiting for real providers
                    if allow_await_cooldown:
                        self._await_cooldown(provider_key, timeout=timeout)
                    if self._is_provider_unusable(provider_key):
                        raise ValueError(
                            f"Provider '{provider_key}' is on cooldown and await_cooldown is false"
                        )

                    # Apply round-robin API key if multiple keys available
                    if provider_instance.has_multiple_keys():
                        provider_instance.get_next_api_key()

                    # Store in local _real_providers cache for this instance as well
                    self._real_providers[provider_key] = provider_instance
                    return provider_instance

                # Otherwise, it's the first time this built-in provider is requested globally
                # Create it and store it in the global cache.
                settings = BUILTIN_PROVIDERS[provider_key]

                # Support multiple API keys: check for both *_API_KEYS and *_API_KEY
                api_key_env = settings["api_key_env"]
                api_keys_env = (
                    api_key_env + "S"
                )  # e.g., OPENAI_API_KEY -> OPENAI_API_KEYS

                # Priority: *_API_KEYS takes precedence over *_API_KEY
                api_keys_value = os.getenv(api_keys_env)
                api_key_value = os.getenv(api_key_env)

                api_keys_list = []
                if api_keys_value:
                    api_keys_list = _parse_api_keys(api_keys_value)
                elif api_key_value:
                    api_keys_list = _parse_api_keys(api_key_value)

                if not api_keys_list:
                    env_var_names = (
                        [api_keys_env, api_key_env]
                        if api_keys_value is None
                        else [api_keys_env]
                    )
                    raise ValueError(
                        f"Built-in provider '{provider_key}' requires "
                        f"one of the environment variables {env_var_names} to be set."
                    )

                if not model_name_for_request:
                    model_name_for_request = settings["default_model"]

                # Check if this built-in provider has already been initialized (e.g., from a prior run of the app)
                # This is important for the singleton pattern to ensure built-ins aren't duplicated or re-validated.
                if provider_key in _GLOBAL_BUILTIN_PROVIDERS:
                    return _GLOBAL_BUILTIN_PROVIDERS[provider_key]

                # Otherwise, proceed with creating and caching the built-in provider
                builtin_config = LLMProviderConfig(
                    name=provider_key,
                    base_url=settings["base_url"],
                    model=model_name_for_request,
                    api_key=api_keys_list[
                        0
                    ],  # Use the first key as the default for Pydantic
                    api_keys=api_keys_list,  # Pass all keys for model_validator
                    temperature=settings.get("temperature", 0.7),
                    max_tokens=settings.get("max_tokens", 4096),
                )
                with _GLOBAL_BUILTIN_LOCK:
                    _GLOBAL_BUILTIN_PROVIDERS[provider_key] = builtin_config
                return builtin_config

        # Original logic for existing providers (real or virtual) from borg.yaml
        if name in self._real_providers:
            # Handle rate limiting for real providers
            if allow_await_cooldown:
                self._await_cooldown(name, timeout=timeout)
            if self._is_provider_unusable(name):
                raise ValueError(
                    f"Provider '{name}' is on cooldown and await_cooldown is false"
                )

            # Get the provider and apply round-robin API key if multiple keys available
            provider = self._real_providers[name]
            if provider.has_multiple_keys():
                provider.get_next_api_key()
            return provider
        elif name in self._virtual_providers:
            return self._get_from_virtual_provider(
                name, approximate_tokens, timeout, allow_await_cooldown
            )
        else:
            # Fallback to check built-in providers if not found in real/virtual and not a specific 'builtin:' request
            # This handles cases where a user might just ask for "openai" without "builtin:openai/..."
            if name in BUILTIN_PROVIDERS:
                # If it's a known built-in, redirect to get it via the built-in mechanism
                return self.get(
                    f"builtin:{name}/{BUILTIN_PROVIDERS[name]['default_model']}"
                )
            raise ValueError(f"LLM provider '{name}' not found")

    def _await_cooldown(
        self, provider_name: str, interval: float = 1.0, timeout: Optional[float] = None
    ):
        with self._lock:
            if provider_name in self._unusable_providers:
                cooldown_end = self._unusable_providers[provider_name]
                current_time = time.time()
                if current_time < cooldown_end:
                    time_to_wait = cooldown_end - current_time
                    if timeout is not None and time_to_wait > timeout:
                        raise TimeoutError(
                            f"Timeout waiting for provider {provider_name} to exit cooldown"
                        )
                    logger.info(
                        f"Provider '{provider_name}' in cooldown. Waiting "
                        f"{time_to_wait:.2f} seconds..."
                    )
                    time.sleep(time_to_wait)
                    # After waiting, re-check, and if still in cooldown (e.g. another thread signaled), raise
                    if self._is_provider_unusable(
                        provider_name
                    ):  # Re-check after sleep
                        raise ValueError(
                            f"Provider '{provider_name}' is still in cooldown after waiting."
                        )
                # If we reached this point, either cooldown has expired or we waited it out.
                # Remove from unusable_providers.
                del self._unusable_providers[provider_name]

    def _get_from_virtual_provider(
        self,
        virtual_provider_name: str,
        approximate_tokens: Optional[int],
        timeout: Optional[float],
        allow_await_cooldown: bool,
    ) -> LLMProviderConfig:
        virtual_config = self._virtual_providers[virtual_provider_name]
        start_time = time.time()

        while True:
            all_resolved_upstreams: List[LLMProviderConfig] = []
            min_cooldown_end_time = float("inf")

            for upstream_info in virtual_config.upstreams:
                upstream_name = upstream_info["name"]

                # If the upstream is currently unusable, update the earliest cooldown time
                if self._is_provider_unusable(upstream_name):
                    current_cooldown_end = self._unusable_providers.get(
                        upstream_name, float("inf")
                    )
                    min_cooldown_end_time = min(
                        min_cooldown_end_time, current_cooldown_end
                    )
                    continue  # Skip this upstream for current attempt

                try:
                    # Recursively call get, allowing nested virtual providers to do their own round-robin
                    resolved_provider = self.get(
                        upstream_name,
                        approximate_tokens,
                        timeout=None,
                        allow_await_cooldown=False,
                    )
                    all_resolved_upstreams.append(resolved_provider)
                except ValueError as e:
                    # If an upstream (real or virtual) is explicitly unusable or fails to provide an LLM, skip it.
                    pass

            if all_resolved_upstreams:
                # Filter resolved upstreams by approximate_tokens
                filtered_upstreams: List[LLMProviderConfig] = []
                if approximate_tokens is not None:
                    for provider in all_resolved_upstreams:
                        if approximate_tokens <= provider.max_tokens:
                            filtered_upstreams.append(provider)
                else:
                    filtered_upstreams = all_resolved_upstreams

                if filtered_upstreams:
                    # Use random selection for virtual provider upstreams to ensure fair distribution
                    # across all possible providers over multiple calls
                    import random

                    selected_provider = filtered_upstreams[0]

                    # Apply round-robin API key if the selected provider has multiple keys
                    if selected_provider.has_multiple_keys():
                        selected_provider.get_next_api_key()
                    return selected_provider

            # If no usable upstreams, determine next action based on allow_await_cooldown and timeout
            if not allow_await_cooldown:
                raise ValueError(
                    f"No eligible upstream providers for virtual provider {virtual_provider_name}. All are on cooldown."
                )

            if min_cooldown_end_time == float("inf"):
                # This should ideally not happen if there are upstreams at all, as min_cooldown_end_time would be set.
                # It implies no upstreams were found or all failed for non-429 reasons (which are already caught above).
                raise ValueError(
                    f"No upstreams found for virtual provider {virtual_provider_name} to await."
                )

            time_to_wait = min_cooldown_end_time - time.time()
            if (
                time_to_wait <= 0
            ):  # Should be handled by _is_provider_unusable, but a safeguard
                continue  # Retry immediately if cooldown already expired

            if timeout is not None:
                time_elapsed = time.time() - start_time
                remaining_timeout = timeout - time_elapsed
                if remaining_timeout <= 0:
                    # Find the provider that will be available soonest
                    earliest_provider = None
                    earliest_time = float("inf")
                    for upstream_info in virtual_config.upstreams:
                        upstream_name = upstream_info["name"]
                        if upstream_name in self._unusable_providers:
                            cooldown_end = self._unusable_providers[upstream_name]
                            if cooldown_end < earliest_time:
                                earliest_time = cooldown_end
                                earliest_provider = upstream_name

                    if earliest_provider:
                        raise TimeoutError(
                            f"Timeout waiting for provider {earliest_provider} to exit cooldown."
                        )
                    else:
                        raise ValueError(
                            f"Timeout of {timeout} seconds reached while waiting for usable upstreams for virtual provider {virtual_provider_name}."
                        )
                # Wait for the minimum of time_to_wait and remaining_timeout
                sleep_duration = min(time_to_wait, remaining_timeout)
                time.sleep(sleep_duration)
            else:
                time.sleep(time_to_wait)
